"""Module for managing rate limits."""

import time


class RateLimiter:
    """
    Limit events to a certain rate, but allow building up
    "credits" for fast bursts.
    """

    def __init__(self, hz: float, burst: float = 250) -> None:
        """The rate limiter will begin with accum_limit credits and
        cannot go above that.  rate is in number of credits added per second.
        """
        self.rate = hz
        # The max number of credits we can accumulate
        self.accum_limit = burst
        # The current number of credits
        self.current_limit = burst

        # The last time we were called
        # and calculated credits
        self.timestamp = time.monotonic()

    def limit(self) -> float:
        """If it hasn't been called too often, subtract one
        credit and return number of credits remaining, Otherwise return 0.

        Credits refill at "rate" per second up to a max of accum_limit
        """
        elapsed = time.monotonic() - self.timestamp

        # Monotonic can go backwards with time travel testing
        elapsed = max(0.0, elapsed)
        self.current_limit += self.rate * elapsed
        self.current_limit = min(self.current_limit, self.accum_limit)
        self.timestamp = time.monotonic()
        if self.current_limit >= 1:
            self.current_limit -= 1
            return self.current_limit

        return 0.0



# SPDX-FileCopyrightText: Copyright Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

from socket import timeout
import threading
import weakref
import logging
import time
import uuid
import traceback
import json
import asyncio

from scullery import messagebus, workers, util

try:
    import msgpack
except Exception:
    pass

logger = logging.getLogger("system.mqtt")


connections = {}

lock = threading.RLock()

# {server, topic, qos : weakref(f)}

allSubscriptions = {}


# list them by local name.  But only real, phsyical MQTT connections.
# This is how passive connections find their real one
connectionsByBusName = weakref.WeakValueDictionary()


def getWeakrefHandlers(self):
    self = weakref.ref(self)

    def on_connect(client, userdata=None, flags=None, rc=0, *a):
        if not rc == 0:
            self().on_disconnected()
            return

        logger.info(
            "Connected to MQTT server: " + self().server + "result code " + str(rc)
        )
        self().on_still_connected()
        # Don't block the network thread too long

        def subscriptionRefresh():
            try:
                with self().lock:
                    for i in allSubscriptions:
                        # Here the bus prefix is our connection ID
                        if i[0] == self().busPrefix:
                            if allSubscriptions[i]():
                                # Refresh all subscriptions
                                self().connection.subscribe(i[1], i[2])
            except Exception:
                logger.exception("Error subscription refresh")

        workers.do(subscriptionRefresh)

    def on_disconnect(client, *a):
        if not self():
            return
        logger.info("Dis_connected from MQTT server: " + self().server)
        self().on_disconnected()
        logger.info("Dis_connected from MQTT server: " + self().server)

    def on_message(client, userdata, msg):
        try:
            s = self()
            # Everything must be fine, because we are getting messages
            messagebus.post_message(
                s.busPrefix + "/in/" + msg.topic,
                msg.payload,
            )
            s.on_still_connected()
        except Exception:
            print(traceback.format_exc())

    return on_connect, on_disconnect, on_message


# Used to fake a crash fro unit testing purposes


class Connection:
    def __init__(
        self,
        server,
        port=1883,
        password=None,
        message_bus_name=None,
        *,
        alert_priority="info",
        alert_ack=True,
        connection_id=""
    ):
        # connection_id is used to ensure separate *physical* connections and prevent reuse
        self.server = server
        self.port = port
        self.lock = threading.Lock()
        self.password = password

        self.message_bus_name = message_bus_name
        self.connection = None

        # Defensive against None
        connection_id = connection_id or ""

        self.is_connected = False

        if not server:
            passive = True
        else:
            passive = False

        self.passive = passive

        server = server or str(uuid.uuid4())
        virtual = server.startswith("__virtual__")

        if passive and (not message_bus_name):
            raise ValueError(
                "No server specified. To create a passive connection you must specify an internal message_bus_name"
            )
        self.busPrefix = "/mqtt/" + server + \
            ":" + str(port) + (connection_id or "")

        self.subscriptions = {}
        logger.info("Creating connection object to: " + self.server)

        self.localStatusTopic = self.busPrefix + "/connectionStatus"

        # paho requires non-python stuff.
        try:
            import paho.mqtt.client as mqtt

            paho = True

        except ImportError:
            logging.exception(
                "PahoMQTT not installed. No MQTT connection possible here. continuing with dummy."
            )
            paho = False

        # When we wrap a function store a weakref to the original here,
        # Pplus the wrapper, so the wrapper doesn't get GCed till
        # The wearkref callback deletes it.
        self.subscribeWrappers = {}

        with lock:
            if connection_id:
                connection_id = "?" + connection_id
            self.connection_id = connection_id

            n = server + ":" + str(port) + connection_id
            if n in connections and connections[n]():
                raise RuntimeError("There is already a connection")
            torm = []
            for i in connections:
                if not connections[i]():
                    torm.append(i)
            for i in torm:
                del connections[i]
            connections[n] = weakref.ref(self)

            if message_bus_name:
                self.busPrefix = "/mqtt/" + message_bus_name

            try:
                if not virtual and not passive:

                    def out_handler(topic, message, timestamp, annotation):
                        self.connection.publish(
                            topic[len(self.busPrefix + "/out/"):],
                            payload=message,
                            qos=annotation[0],
                            retain=annotation[1],
                        )

                else:
                    if not passive:
                        # Virtual loopback server doesn't actually use a real server
                        def out_handler(topic, message, timestamp, annotation):
                            t = topic[len(self.busPrefix + "/out/"):]
                            messagebus.post_message(
                                self.busPrefix + "/in/" + t, message)

                if not passive:
                    messagebus.subscribe(
                        self.busPrefix + "/out/#", out_handler)
                    self.out_handler = out_handler

                if not virtual and not passive:
                    x = server.split("@")

                    if len(x) == 2:
                        host = x[1]
                        user = x[0]
                    elif len(x) == 1:
                        host = x[0]
                        user = None
                    else:
                        raise ValueError(
                            "More than one @ symbol in server name??")

                    self.username = user
                    self.password = password

                    self.configure_alert(alert_priority, alert_ack)

                else:
                    self.connection = None
                    self.configure_alert(alert_priority, alert_ack)
                    self.on_still_connected()

                if not passive:
                    connectionsByBusName[self.busPrefix] = self

            except Exception:
                # Attempt cleanup
                try:
                    self.connection.disconnect()
                    time.sleep(2)
                except Exception:
                    pass
                try:
                    del connections[server + ":" + str(port)]
                except Exception:
                    pass
                raise

    def on_connection_crash(self, tb):
        print(traceback.format_exc())
        self.reconnect()

    def on_still_connected(self):
        if not self.is_connected:
            messagebus.post_message(self.localStatusTopic, "connected")
        self.is_connected = True
        pass

    def on_disconnected(self):
        logging.warning(
            "A connection has dis_connected from MQTT server: " + self.server
        )
        if self.is_connected:
            messagebus.post_message(self.localStatusTopic, "dis_connected")
        self.is_connected = False

    def subscribe_to_status(self, f):
        messagebus.subscribe(self.localStatusTopic, f)

    def close(self):
        # Attempt cleanup
        try:
            self.connection.disconnect()
        except Exception:
            pass
        try:
            del connections[self.server + ":" +
                            str(self.port) + self.connection_id]
        except Exception:
            pass

    def __del__(self):
        if self.connection:
            self.connection.disconnect()

    def unsubscribe(self, topic, function):
        with self.lock:
            # Very expensive to search this dict like this, but unsub shouldn't happen much.
            try:
                torm = []
                # Find any subscriptions to the topic for this particular bus prefix and delete them
                for i in allSubscriptions:
                    if (
                        i[0] == self.busPrefix
                        and i[1] == topic
                        and allSubscriptions[i]() == function
                    ):
                        torm.append(i)
                for i in torm:
                    allSubscriptions.pop(i)
            except KeyError:
                print(traceback.format_exc())
                pass

            torm = []
            for i in self.subscribeWrappers:
                x = self.subscribeWrappers[i]
                if x[2] == topic and (x[0]() == function or x[0]() is None):
                    messagebus.unsubscribe(x[3], x[1])
                    torm.append(i)
            for i in torm:
                try:
                    del self.subscribeWrappers[i]
                except KeyError:
                    pass

            for i in self.subscribeWrappers:
                x = self.subscribeWrappers[i]
                if x[2] == topic:
                    return

            # We could not find even a single subscriber function
            # So we unsubscribe at the MQTT level
            logging.debug("MQTT Unsubscribe from " +
                          topic + " at " + self.server)
            if self.connection:
                self.connection.unsubscribe(topic)

    def configure_alert(self, *a):
        pass

    def _mqttSubscribe(topic, qos):
        if self.connection:
            self.connection.subscribe(topic, qos)

    def subscribe(self, topic, function, qos=2, encoding="json"):
        with lock:
            if self.connection:
                self.connection.subscribe(topic, qos)
            else:
                # We are a "Passive" connecytion relaying through a real connection elsewhere in code.
                if self.passive:
                    if self.busPrefix in connectionsByBusName:
                        try:
                            backend = connectionsByBusName[self.busPrefix]
                        except KeyError:
                            backend = None

                        if backend:
                            # The backend to relay through actually exists! So we have to actually tell them to physically
                            # subscribe to the MQTT topic.

                            # if they don't exist yet that is fine, they will look through the subscriptions master list and reconnect anything active anyway
                            backend.connection.subscribe(topic, qos)

        x = str(uuid.uuid4())

        def handleDel(*a):
            try:
                del self.subscribeWrappers[x]
            except KeyError:
                pass
            # We're really just using the "check if there's no subscribers"
            # Part of the function
            self.unsubscribe(topic, None)

        fID = id(function)
        function = util.universal_weakref(function, handleDel)

        # This is our master list of who is subscribed where. It is used for reconnection.
        # It is also used so that subscriptions can persist beyond any given connection object!!!
        # use fID in case of two subscribers to one topic needing to fit in one dict.
        with self.lock:
            allSubscriptions[self.busPrefix, topic, qos, fID] = function

        # Connection.subscribe was blocking forever.
        # Use a different thread, to hopefully avoid deadlocks

        def backgroundSubscribeTask():
            with self.lock:
                self.connection.subscribe(topic, qos)

        if encoding == "json":

            def wrapper(t, m):
                # Get rid of the extra kaithem framing part of the topic
                t = t[len(self.busPrefix + "/in/"):]
                if not isinstance(m, str):
                    m = m.decode("utf-8")
                try:
                    m = json.loads(m)
                except Exception:
                    logging.debug("Bad JSON:" + m[:64])
                function()(t, m)

        elif encoding == "msgpack":

            def wrapper(t, m):
                # Get rid of the extra kaithem framing part of the topic
                t = t[len(self.busPrefix + "/in/"):]
                function()(t, msgpack.unpackb(m, raw=False))

        elif encoding == "utf8":

            def wrapper(t, m):
                # Get rid of the extra kaithem framing part of the topic
                t = t[len(self.busPrefix + "/in/"):]
                if not isinstance(m, str):
                    m = m.decode("utf-8")
                function()(t, m)

        elif encoding == "raw":

            def wrapper(t, m):
                # Get rid of the extra kaithem framing part of the topic
                t = t[len(self.busPrefix + "/in/"):]
                function()(t, m)

        else:
            raise ValueError("Invalid encoding: " + encoding)

        # Correctly call message bus error handlers
        wrapper.messagebusWrapperFor = function

        internalTopic = self.busPrefix + "/in/" + topic

        # Extra data is mostly used for unsubscription
        self.subscribeWrappers[x] = (function, wrapper, topic, internalTopic)

        # logging.debug("MQTT subscribe to " + topic + " at " + self.server)
        # Ref to f exists as long as the original does because it's kept in subscribeWrappers
        messagebus.subscribe(internalTopic, wrapper)

        # Important we do this *after*, because the server might auto-send us something that could be important for.
        # Only the first function will get it, but whatever, at least we can see certain things when manually subscribing, for test purposes.
        # Or should we do it before, to resolve that bit of ambiguity and never get the message???
        if self.connection:
            workers.do(backgroundSubscribeTask)

    def publish(self, topic, message, qos=2, encoding="json", retain=False):
        if encoding == "json":
            message = json.dumps(message)
        elif encoding == "msgpack":
            message = msgpack.packb(message, use_bin_type=True)
        elif encoding == "utf8":
            message = message.encode("utf8")
        elif encoding == "raw":
            pass
        else:
            raise ValueError("Invalid encoding!")
        messagebus.post_message(
            self.busPrefix + "/out/" + topic, message, annotation=(qos, retain)
        )


def get_connection(
    server,
    port=1883,
    password=None,
    message_bus_name=None,
    *,
    alert_priority="info",
    alert_ack=True,
    connection_id=""
):
    # Kaithem is gonna monkeypatch kaithem with one that has better
    # logging
    global Connection

    # Blank password means no password
    password = password or None
    with lock:
        x = None

        connection_idSuffix = ""
        if connection_id:
            connection_idSuffix = "?" + connection_id

        if server + ":" + str(port) + connection_idSuffix in connections:
            x = connections[server + ":" + str(port) + connection_idSuffix]()

        if x:
            if not message_bus_name == x.message_bus_name:
                # We can safely use the existing one.   If t
                logging.warning(
                    "Using connection that already exists, but with a different message bus name:  "
                    + x.message_bus_name
                )

                message_bus_name = x.message_bus_name

            if x.password or password:
                if not x.password == password:
                    raise ValueError(
                        "There is already a connection to the same host and user, but with a different password."
                    )

            x.configure_alert(alert_priority, alert_ack)
            return x

        return Connection(
            server,
            port,
            password=password,
            alert_ack=True,
            alert_priority="info",
            message_bus_name=message_bus_name,
            connection_id=connection_id,
        )



# SPDX-FileCopyrightText: Copyright Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later


import sys
import os
import weakref
import types
from typing import TypeVar
# Credit to Jay of stack overflow for this function


def which(program):
    "Check if a program is installed like you would do with UNIX's which command."

    # Because in windows, the actual executable name has .exe while the command name does not.
    if sys.platform == "win32" and not program.endswith(".exe"):
        program += ".exe"

    # Find out if path represents a file that the current user can execute.
    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    # If the input was a direct path to an executable, return it
    if fpath:
        if is_exe(program):
            return program

    # Else search the path for the file.
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    # If we got this far in execution, we assume the file is not there and return None
    return None


T = TypeVar("T")


def universal_weakref(f: T, cb=None) -> weakref.ref[T]:
    "Return a weakref or weakmethod as appropriate for f"
    if isinstance(f, types.MethodType):
        # TODO why doesn't linter know this is safe?
        return weakref.WeakMethod(f, cb)  # type: ignore
    else:
        return weakref.ref(f, cb)


def search_paths(fn: str, paths: list[str]) -> str | None:
    for i in paths:
        if os.path.exists(os.path.join(i, fn)):
            return os.path.join(i, fn)



# SPDX-FileCopyrightText: Copyright Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

import stat
import io
import sys
import os
import threading
import gzip
import bz2
import json
import time
import logging
import hashlib
import threading

posix_rename = False
if sys.platform.startswith("linux"):
    posix_rename = True
if sys.platform.startswith("darwin"):
    posix_rename = True


def resolve_path(fn, expand=True):
    if not expand:
        return fn
    return os.path.expandvars(os.path.expanduser(fn))


# TODO: Ensure only one thread can save a file at a time
strio = io.BytesIO

# must be Rlock because load is recursive
lock = threading.RLock()


def chmod_private_try(p, execute=True):
    try:
        if execute:
            os.chmod(p, stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IXGRP)
        else:
            os.chmod(p, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP)
    except Exception as e:
        raise e


def ensure_dir(f):
    d = os.path.dirname(f)
    if not os.path.exists(d):
        os.makedirs(d)


def save(data, fn, *, private=False, backup=True, expand=True, md5=False, nolog=False):
    """Save data to file. Filename must end in .json, .yaml, .txt, or .bin. Data will be encoded appropriately.
    Also supports compressed versions via filenames ending in .gz or .bz2.
    Args:
        data:
            the data to be written. if fn is a .json or .yaml, must be serializable. If filename is .txt, must be a string.
            If .bin, must be something like bytes.
        private:
            If True, file created with mode 700(Full access to root and owner but not even read to anyone else)
            If False(the default), file created with default mode
        backup:
            Setting this to true is an alias for mode="backup"
    """
    fn = resolve_path(fn, expand)
    with lock:
        # Make sure we don't overwrite a file when we create our dirs, because that behavior is undocumented in makedirs.
        x = os.path.split(fn)[0]
        already = {}

        # Safety counter to stop really wierd loops
        for i in range(64):
            if os.path.isfile(x):
                raise RuntimeError("Required intermediate directory is already present as a file, refusing to overwrite file")
            x = os.path.split(x)[0]
            # Loop prevention
            if x in already:
                break
            already[x] = True

        if not os.path.exists(os.path.dirname(fn)):
            if private:
                os.makedirs(os.path.dirname(fn), mode=0o700)
            else:
                os.makedirs(os.path.dirname(fn))

        if os.path.isdir(fn):
            raise RuntimeError("Filename is already present as a directory, refusing to overwrite directory")
        # Get base type without compression
        if fn.endswith(".gz"):
            x = fn[:-3]
        elif fn.endswith(".bz2"):
            x = fn[:-4]
        else:
            x = fn
        # Encode the data into our chosen format
        if x.endswith(".json"):
            data = json.dumps(data).encode("utf8")
        elif x.endswith(".yaml"):
            import yaml

            data = yaml.dump(data).encode("utf8")

        elif x.endswith(".toml"):
            import toml

            data = toml.dumps(data).encode("utf8")

        elif x.endswith((".txt", ".md", ".rst")):
            data = str(data).encode("utf8")
        elif x.endswith(".bin"):
            data = data
        else:
            raise ValueError("Unsupported or missing File Extension")

        # We have selected a compressed type. Compress in-memory first so we can read-before-write
        # Note that disk access is slow enough the call to  basically makes no difference in speed here if it's already imported
        if fn.endswith(".gz"):
            i = strio()
            f = gzip.GzipFile(fn, mode="wb", fileobj=i)
            f.write(data)
            f.close()
            data = i.getvalue()

        elif fn.endswith(".bz2"):
            c = bz2.BZ2Compressor()
            c.compress(data)
            data = c.flush()
            del c

        # Do a read-before-write. We don't write if we don't have to
        if os.path.exists(fn):
            with open(fn, "rb") as f:
                if f.read() == data:
                    return

        ensure_dir(os.path.split(fn)[0])

        if backup:
            tempfn = fn + str(time.time())
        else:
            tempfn = fn

        if not nolog:
            logging.debug("Writing: " + fn)
        # Actually write it
        with open(tempfn, "wb") as f:
            # In backup mode, pre truncate and flush.
            # This means that even if an error occors during writing, we will be able to tell by the
            # mtime that the tilde file is the right one.
            if backup:
                f.flush()
                os.fsync(f.fileno())

            # Chmod it before we write anything to it.
            if private:
                chmod_private_try(tempfn)
            f.write(data)
            f.flush()
            os.fsync(f.fileno())

        if backup:
            os.replace(tempfn, fn)

        if md5:
            with open(fn + ".md5", "w") as md5f:
                md5f.write(hashlib.md5(data).hexdigest())


def load(filename, *, expand=True):
    """Load a file. Return str if file extension is .txt, bytes on .bin, dict on .yaml or .json.

    After that may be a .bz2 or a .gz for compression.

    """
    filename = resolve_path(filename, expand)

    with lock:
        try:
            # Open the file and get the filename without the compression type attached to it.
            if filename.endswith(".gz"):
                f = gzip.GzipFile(filename, mode="rb")
                x = filename[:-3]
            elif filename.endswith(".bz2"):
                x = filename[:-4]
                f = bz2.BZ2File(filename, mode="rb")
            else:
                f = open(filename, "rb")
                x = filename

            if x.endswith(".json"):
                r = json.loads(f.read().decode("utf8"))
            elif x.endswith(".yaml"):
                import yaml

                r = yaml.load(f.read().decode("utf8"), Loader=yaml.SafeLoader)

            elif x.endswith(".toml"):
                import toml

                r = toml.loads(f.read().decode("utf8"))

            elif x.endswith((".txt", ".md", ".rst")):
                r = f.read().decode("utf8")
            elif x.endswith(".bin"):
                r = f.read()
            else:
                raise ValueError("Unsupported File Extension")
        except Exception as e:
            try:
                f.close()
            except Exception:
                pass
            raise
        try:
            f.close()
        except Exception:
            pass

        return r



# SPDX-FileCopyrightText: Copyright Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
This file manages global message bus.  Think of it as an internal version of MQTT.
In fact, the same wildcard syntax as MQTT is supported.

The global subscribe, unsubscribe, and post_message functions run subscribers in
a background thread by default.

"""

import weakref
import threading
import time
import traceback
import logging
import inspect
import types
import copy
from typing import Any
from collections.abc import Callable

import beartype

from . import workers
from collections import defaultdict, OrderedDict

_subscribers_list_modify_lock = threading.RLock()
cachelock = threading.RLock()
# OrderedDict doesn't seem as fast as dict. So I have a cache of the cache
parsecache = OrderedDict()
parsecachecache = {}

log = logging.getLogger("system.messagebus")


def normalize_topic(topic: str) -> str:
    """ "Because some topics are equivalent("/foo" and "foo"), this lets us convert them to the canonical "/foo" representation.
    Note that "/foo/" is not the same as "/foo", because a trailing slash indicates a "directory"."""
    topic = topic.strip()
    if not topic.startswith("/"):
        return "/" + topic
    else:
        return topic


def _shouldReRaiseAttrErr():
    return True


subscriber_error_handlers = []


def _handle_error(f: Callable[..., Any], topic: str, value: Any):
    log.exception("Message bus subscriber error")
    if hasattr(f, "messagebusWrapperFor"):
        f = f.messagebusWrapperFor
    for i in subscriber_error_handlers:
        try:
            i(f, topic, value)
        except Exception:
            print(traceback.format_exc())


def _run_function(f, a):
    return f(*a)


class MessageBus:
    def __init__(self, executor: Callable[..., Any] | None = None):
        """You pass this a function of one argument that just calls its argument.
        Defaults to calling in same thread and ignoring errors.

        If you pass it an executor, the executor must take a
        callable and call it in a background thread.

        """
        if executor == None:

            def do(self, f: Callable[[], Any]):
                try:
                    f()
                except Exception:
                    pass

            self._executor = do
        else:
            self._executor = executor

        self._subscribers = defaultdict(list)
        self._subscribers_immutable = {}

    @beartype.beartype
    def subscribe(self, topic: str, callback: Callable[..., Any]):
        topic = normalize_topic(topic)

        with _subscribers_list_modify_lock:
            wrappedCallback = self._wrap_callback(callback, topic)

            self._subscribers[topic].append(wrappedCallback)
            self._subscribers_immutable = copy.deepcopy(self._subscribers)

    def unsubscribe(self, topic: str, function: Callable[..., Any]):
        "Unsubscribe topic from function"
        try:
            with _subscribers_list_modify_lock:
                target = None
                for j in self._subscribers[topic]:
                    if j.originalFunction() == function:
                        target = j
                if target:
                    self._subscribers[topic].remove(target)
                else:
                    pass
        except Exception:
            print(traceback.format_exc())
        # There is a very slight chance someone will
        # Add something to topic before we delete it but after the test.
        # That would result in a canceled subscription
        # So we use this lock.
        try:
            with _subscribers_list_modify_lock:
                if not self._subscribers[topic]:
                    self._subscribers.pop(topic)
        except AttributeError as e:
            # This try and if statement are supposed to catch nuisiance errors when shutting down.
            if _shouldReRaiseAttrErr():
                raise e

        finally:
            with _subscribers_list_modify_lock:
                self._subscribers_immutable = copy.deepcopy(self._subscribers)

    @staticmethod
    def parse_topic(topic: str) -> set[str]:
        """Parse the topic string into a list of all the different subscriptions
        that could possibly match, including wildcards"""
        global parsecache
        # Since this is a pure function(except the caching itself) we can cache it
        if topic in parsecache:
            return parsecache[topic]

        # Let's cache by the original version of the topic, so we don't have to convert it to the canonical
        oldtopic = topic
        topic = normalize_topic(topic)

        # A topic foo/bar/baz would go to
        # foo, foo/bar, and /foo/bar/baz
        # So we need to make a list like that
        matchingtopics = {"/#"}
        parts = topic.split("/")
        last = ""

        # Add the exact one
        matchingtopics.add(topic)

        for i in parts:
            last += i + "/"
            matchingtopics.add(last + "#")
        parsecache[oldtopic] = matchingtopics
        # Don't let the cache get too big.
        # Getting rid of the oldest should hopefully converge to the most used topics being cached
        if len(parsecache) > 600:
            parsecache.popitem(last=False)
        return matchingtopics

    @beartype.beartype
    def _wrap_callback(self, f: Callable[..., Any], topic: str):
        """return function g that calls f with (topic,message) or just f(topic), depending
        on how many args there are.
         and if errors is true logs the error"""

        args = len(inspect.signature(f).parameters)

        timestamp = time.time()

        try:
            desc = str(f.__name__ + " of " + f.__module__)
        except Exception:
            desc = str(f)

        # Allright, here is how this works.
        # We have to deal with the possibility that, at any time,
        # The callback will cease to exist. That, in fact, is how one unsubscribes.
        # So, we make this here closure that knows the topic, and
        # When the GC goes Om Nom Nom on the function, we get passed the weakref to it.
        # Then we get rid of the empty weakref and if that causes the entire topic
        # To have no subscribers, delete that too in case of memory leak.

        def delsubscription(weakrefobject):
            if time.time() < timestamp - 0.5:
                logging.warning("Function: " + desc + " was deleted 0.5s after being subscribed.  This is probably not what you wanted.")

            try:
                with _subscribers_list_modify_lock:
                    self._subscribers[topic].remove(weakrefobject)
            except Exception:
                pass
            # There is a very slight chance someone will
            # Add something to topic before we delete it but after the test.
            # That would result in a canceled subscription
            # So we use this lock.
            try:
                with _subscribers_list_modify_lock:
                    if not self._subscribers[topic]:
                        self._subscribers.pop(topic)
            except AttributeError as e:
                # This try and if statement are supposed to catch nuisiance errors when shutting down.
                if _shouldReRaiseAttrErr():
                    raise e

            finally:
                with _subscribers_list_modify_lock:
                    self._subscribers_immutable = copy.deepcopy(self._subscribers)

        if isinstance(f, types.MethodType):
            f = weakref.WeakMethod(f, delsubscription)
        else:
            f = weakref.ref(f, delsubscription)

        # Mutable object that gets saved to the closure for keeping track of if we already loggged this
        alreadyLogged = [False]
        if args == 0:

            def g(topic, message, errors, timestamp, annotation):
                try:
                    f2 = f()
                    if f2:
                        f2()
                except Exception:
                    try:
                        if errors:
                            if not alreadyLogged[0]:
                                global _handle_error
                                _handle_error(f2, topic, message)
                            alreadyLogged[0] = True
                    except Exception as e:
                        print("err", e)
        elif args == 2:

            def g(topic, message, errors, timestamp, annotation):
                try:
                    f2 = f()
                    if f2:
                        f2(topic, message)
                except Exception:
                    try:
                        if errors:
                            if not alreadyLogged[0]:
                                global _handle_error
                                _handle_error(f2, topic, message)
                            alreadyLogged[0] = True

                    except Exception as e:
                        print("err", e)
        elif args == 4:

            def g(topic, message, errors, timestamp, annotation):
                try:
                    f2 = f()
                    if f2:
                        f2(topic, message, timestamp, annotation)
                except Exception:
                    try:
                        if errors:
                            if not alreadyLogged[0]:
                                global _handle_error
                                _handle_error(f2, topic, message)
                            alreadyLogged[0] = True

                    except Exception as e:
                        print("err", e)
        elif args == 1:

            def g(topic, message, errors, timestamp, annotation):
                try:
                    f2 = f()
                    if f2:
                        f2(message)
                except Exception:
                    try:
                        if errors:
                            if not alreadyLogged[0]:
                                global _handle_error
                                _handle_error(f2, topic, message)
                            alreadyLogged[0] = True
                    except Exception as e:
                        print("err", e)
        else:
            raise ValueError("Invalid function signature(0,1,2, or 4 args supported, not " + str(args) + ")")

        # Ref to the weakref so it's easy to check if the function we are wrapping
        # Still exists.
        g.originalFunction = f
        return g

    def _post(self, topic, message, errors, timestamp, annotation, executor=None):
        executor = executor or self._executor

        matchingtopics = self.parse_topic(topic)
        # We can't iterate on anything that could possibly change so we make copies
        d = self._subscribers_immutable
        for i in matchingtopics:
            if i in d:
                # When we find a match, we make a copy of that subscriber list
                x = d[i][:]
                # And iterate the copy
                for f in x:
                    # we call the ref to get its refferent
                    # An error could happen in the subscriber
                    # Or a typeerror could because the weakref has been collected
                    # We ignore both of these errors and move on
                    executor(f, (topic, message, errors, timestamp, annotation))

    def post_message(
        self,
        topic: str,
        message: Any,
        errors: bool = True,
        timestamp: float | None = None,
        annotation: Any = None,
        synchronous: bool = False,
    ):
        # Use the executor to run the post message job
        # To allow for the possibility of it running in the background as a thread
        global _run_function
        topic = normalize_topic(topic)
        try:
            topic = str(topic)
        except Exception:
            raise TypeError("Topic must be a string or castable to a string.")

        timestamp = timestamp or time.time()
        self._post(
            topic,
            message,
            errors,
            timestamp,
            annotation,
            _run_function if synchronous else None,
        )


# Setup the default system messagebus
_bus = MessageBus(workers.do)
subscribe = _bus.subscribe
unsubscribe = _bus.unsubscribe
post_message = _bus.post_message



# SPDX-FileCopyrightText: Copyright Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

import threading
import traceback
import logging
import time
import collections
import random

from typing import Callable, List, Optional, Any

# I'm reserving the system log for a reasonable low-rate log.
# So this is in a separate namespace that shows up elsewhere.
logger = logging.getLogger("workers")

syslogger = logging.getLogger("system")

lastWorkersError = 0
backgroundFunctionErrorHandlers = []


def testLatency():
    start_time = time.monotonic()
    x = [0]

    def f():
        x[0] = time.monotonic()

    do(f)

    while time.monotonic() - start_time < 10:
        time.sleep(0.0001)
        if x[0]:
            return x[0] - start_time
    raise RuntimeError("No response")


spawnLock = threading.RLock()

maxWorkers = 32
minWorkers = 4

shutdownWait = 60
run = True
taskQueue = collections.deque()


def inWaiting():
    return len(taskQueue)


def waitingtasks():
    "Return the number of tasks in the task queue"
    return len(taskQueue)


def handle_errorInFunction(f):
    print("Error in: " + str(f))


def stop():
    global run
    run = False


def EXIT():
    # Tell all worker threads to stop and wait for them all to finish.
    # If they aren't finished within the time limit, just exit.
    t = time.time()
    stop()
    for i in workers:
        try:
            # All threads total must be finished within the time limit
            workers[i].join(shutdownWait - (time.time() - t))
            # If we try to exit befoe the thread even has time to start or something
        except RuntimeError:
            pass


workers = {}
workersMutable = {}

wakeupHandles = []
wakeupHandlesMutable = []


def testIntegrity():
    with spawnLock:
        assert len(wakeupHandles) == len(workers)


lastStoppedThread = 0


def _makeWorker(e, q, id, fastMode=False):
    # one worker that just pulls tasks from the queue and does them. Errors are caught and
    # We assume the tasks have their own error stuff
    def workerloop():
        global workers
        global lastStoppedThread
        global wakeupHandles

        f = None

        shouldRun = True

        monotonic = time.monotonic

        lastActivity = monotonic()
        pop = taskQueue.pop

        runningState = [True]
        handle = (e, runningState)
        with spawnLock:
            wakeupHandlesMutable.append(handle)
            wakeupHandles = wakeupHandlesMutable[:]

        while run:
            try:
                runningState[0] = True
                # While either our direct  queue or the overflow queue has things in it we do them.
                while len(taskQueue):
                    try:
                        f = pop()
                    except Exception:
                        f = None

                    if f:
                        try:
                            f[0](*f[1])
                            lastActivity = monotonic()
                        except Exception:
                            global lastWorkersError
                            try:
                                if lastWorkersError < monotonic() - 60:
                                    syslogger.exception(
                                        "Error in function. This message is ratelimited, see debug logs for full.\r\nIn "
                                        + f[0].__name__
                                        + " from "
                                        + f[0].__module__
                                        + "\r\n"
                                    )
                                    lastWorkersError = monotonic()

                                logger.exception("Error in function running in thread pool " + f[0].__name__ + " from " + f[0].__module__)
                            except Exception:
                                print("Failed to handle error: " + traceback.format_exc(6))

                            for i in backgroundFunctionErrorHandlers:
                                try:
                                    i(f)
                                except Exception:
                                    print("Failed to handle error: " + traceback.format_exc(6))
                        finally:
                            # We do not want f staying around, if might hold references that should be GCed away immediatly
                            f = None
                # Ensure the lock is acqured so we can sto the next time
                # We try in blocking mode
                e.acquire(False)
                runningState[0] = False

                # This check happens *after* setting e.on false, so that if we
                # set it false right after they checked and put something in
                # the queue we get it next loop
                if not taskQueue:
                    # Randomize, so they don't all sync up
                    # FastMode polls at 100Hz
                    x = e.acquire(timeout=(random.random() * 1) if not fastMode else 0.01)
                    runningState[0] = True

                    if not x:
                        if not shouldRun:
                            return

                        # Fast prelim check.  Allow going below min workers if no activity for 5s
                        # Otherwise stay at min workers
                        if (len(workers) > minWorkers) or (lastActivity < (monotonic() - 5)):
                            # The elements of handle are never copied anywhere,
                            # Once the list is clear, we can be sure there is no further inserts, and the next round will catch almost
                            # all race conditions. Any remaining one in a million ones will be caught in 1 second
                            if lastActivity < (monotonic() - 1):
                                # This should not block.
                                if spawnLock.acquire(timeout=2):
                                    try:
                                        if (len(workers) > minWorkers) or (lastActivity < (monotonic() - 5)):
                                            # Only stop one thread per second to prevent
                                            # chattering

                                            if lastStoppedThread < (monotonic() - 2):
                                                lastStoppedThread = monotonic()
                                                del workersMutable[id]
                                                workers = workersMutable.copy()
                                                wakeupHandlesMutable.remove(handle)
                                                wakeupHandles = wakeupHandlesMutable[:]
                                                return
                                    finally:
                                        spawnLock.release()
            except Exception:
                print("Exception in worker loop: " + traceback.format_exc(6))

    return workerloop


def addWorker():
    global workers
    if spawnLock.acquire(timeout=60):
        try:
            q = []
            e = threading.Lock()
            e.acquire()

            id = time.time()
            # First worker always polls at 100hz
            t = threading.Thread(
                target=_makeWorker(e, q, id),
                name="nostartstoplog.ThreadPoolWorker-" + str(id),
            )
            workersMutable[id] = t
            t.start()
            workers = workersMutable.copy()
        finally:
            spawnLock.release()
    else:
        raise RuntimeError("Could not get the lock!")


_append = taskQueue.append


def do(func: Callable[..., Any], args: Optional[List[Any]] = None):
    """Run a function in the background

    funct(function):
        A function of 0 arguments to be ran in the background in another thread immediatly,
    """

    args = args or []
    if not callable(func):
        raise ValueError("Non callable value")

    _append((func, args))

    for i in wakeupHandles:
        try:
            if i[0].locked():
                i[0].release()
                return
        except RuntimeError:
            pass

    if len(workers) > 4:
        # Wait and retry before attempting to spawn a new thread, if there is probable already enough.
        # that is a slow problematic thing
        time.sleep(0.001)
        time.sleep(0.0001)

        for i in wakeupHandles:
            try:
                if i[0].locked():
                    i[0].release()
                    return
            except RuntimeError:
                pass

    # Sleep 1/25000th of a second for every item in the queue past the max number of threads
    # In an attempt to rate limit
    if len(taskQueue) > maxWorkers:
        time.sleep(max(0, (len(taskQueue) - maxWorkers) / 25000))

    # No unbusy threads? It must go in the overflow queue.
    # Soft rate limit here should work a bit better than the old hard limit at keeping away
    # the deadlocks.
    # Under lock

    # We also need this fast preliminary check to use that lock as rarely as possible.
    if len(workers) < maxWorkers:
        if spawnLock.acquire(timeout=15):
            try:
                if len(workers) < maxWorkers:
                    addWorker()
                    return
            finally:
                spawnLock.release()
        else:
            print("COULD NOT GET SPAWN LOCK TO CREATE THREAD. RESTART SUGGESTED")

    # If we can't spawn a new thread
    # Wait a maximum of 15ms before
    # just giving up and leaving
    # it for when somethin
    # wakes up
    for n in range(25):
        if not taskQueue:
            return
        for i in wakeupHandles:
            try:
                if i[0].locked():
                    i[0].release()
                    return
            except RuntimeError:
                pass
        time.sleep(0.0005)


do_try = do


def start(count=12, qsize=64, shutdown_wait=60):
    "Now just sets parameters, threads are made and destroyed on-demand"
    global __queue, run, workers, maxWorkers
    run = True

    maxWorkers = count



# SPDX-FileCopyrightText: Copyright Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

import math

_unit_registry = None
# Units datafile.  All the units are all mixed together.  Every type of unit has to have a single base unit
# and everything else is defined in terms of that
# On errors we fall back to the Pint library.
units = {
    # Base unit of mass is grams
    "g": 1.0,
    "lb": 453.592,
    "oz": 28,
    # Base unit of distance is meters
    "m": 1.0,
    "in": 0.0254,
    # Base unit of temperature is Kelvin
    "K": 1.0,
    # Offset unit. It is defined by two functions,the one to convert to the base an one to convert from
    "degC": (lambda x: x + 273.15, lambda x: x - 273.15),
    # Voltage/Current
    "A": 1,
    "V": 1,
    # Power, base is W
    "W": 1,
    "dBm": (lambda x: (10 ** (x / 10)) / 1000, lambda x: 10 * math.log10(x / 0.001)),
    # Nobody really uses the uno I don't think but it's the only name we have
    "U": 1,
    "uno": 1,
    "%": 0.01,
    "ppm": 100 / 10**6,
    "ppb": 100 / 10**9,
    "dB": (lambda x: 10 ** (x * 10), lambda x: 10 * math.log10(x)),
    "Pa": 1,
    "psi": 6894.76,
    "PSI": 6894.76,
    "bar": 100000,
    "millibar": 100,
    "boolean": 1,
    "bool": 1,
    "KPH": 1,
    "MPH": 1.60934,
    "km/h": 1,
}

unit_types = {
    # Base unit of mass is grams
    "g": "mass",
    "lb": "mass",
    "oz": "mass",
    "boolean": "boolean",
    "boolean": "boolean",
    "MPH": "speed",
    "KPH": "speed",
    "km/h": "speed",
    # Base unit of distance is meters
    "m": "length",
    "in": "length",
    # Base unit of temperature is Kelvin
    "K": "temperature",
    # Offset unit. To convert FROM the base divide by the first and add the second number
    "degC": "temperature",
    "V": "voltage",
    "A": "current",
    "W": "power",
    "dBm": "power",
    "%": "ratio",
    "ppm": "ratio",
    "ppb": "ratio",
    "U": "ratio",
    "uno": "ratio",
    "dB": "ratio",
    "psi": "pressure",
    "Pa": "pressure",
    "PSI": "pressure",
    "bar": "pressure",
    "millibar": "pressure",
}

_prefixes = {
    "n": 10**-9,
    "u": 10**-6,
    "m": 0.001,
    "k": 1000,
    "M": 1000000,
    "G": 1000000000,
    "T": 1000000000000,
}


def get_unit_type(u):
    return unit_types(u)


def parse_prefix(u):
    if u in units:
        return (units[u], 1)
    elif u[0] in _prefixes and u[1:] in units:
        if not isinstance(units[u[1:]], (int, float)):
            raise ValueError("SI prefix not supported with nonlinear units")
        return (units[u[1:]], _prefixes[u[0]])

    raise KeyError("No such unit: " + u)


def define_unit(unitname, multiplier, type, offset=0, base=None):
    """Define a new unit. The multiplier is how many of the base unit one step in the base unit new unit is worth.
    Base defaults to 1, the global base unit.

    Multiplier can also be a tuple of two functions, frombase,tobase, that handle the conversions.
    Function based units cannot be chained. You must define them in terms of the global default base

    """

    if base:
        z = units[base]
        if isinstance(z, (int, float)):
            multiplier *= z
        else:
            raise RuntimeError("Function-based units can only be defined directly in terms of the global base")

        if not isinstance(multiplier, (int, float)):
            raise RuntimeError("Function-based units can only be defined directly in terms of the global base")

    if isinstance(multiplier, (int, float)):
        multiplier = multiplier
    units[unitname] = multiplier
    unit_types[unitname] = type


define_unit(
    "degF",
    (lambda x: (x + 459.67) * (5 / 9), lambda x: (x * (9 / 5)) - 459.67),
    "temperature",
)

define_unit("ft", 12, "length", 0, "in")
define_unit("mile", 1609.344, "length")

define_unit("m3/min", 1, "flow")
define_unit("cfm", 0.028316847, "flow")
define_unit("gpm", 0.0037854118, "flow")

define_unit("Pa", 1, "pressure")
define_unit("psi", 6894.7573, "pressure")
define_unit("ksi", 1000, "pressure", 0, "psi")
define_unit("mmHg", 133.32237, "pressure")


def _loadPint():
    global _unit_registry
    if not _unit_registry:
        import pint

        _unit_registry = pint.UnitRegistry()


def convert(v: float | int, from_unit: str, to_unit: str) -> float:
    """_summary_

    Args:
        v (float | int): Input value
        from_unit (str): Input Unit
        to_unit (str): Outout unit

    Returns:
        _type_: Value in to_unit
    """
    if from_unit == to_unit:
        return v
    try:
        fr, frmul = parse_prefix(from_unit)
        to, tomul = parse_prefix(to_unit)
    # Fallback to pint library
    except KeyError:
        if not _unit_registry:
            _loadPint()
        x = _unit_registry.Quantity(v, from_unit)
        return x.to(to_unit).magnitude

    v *= frmul
    # Convert into the base unit
    if isinstance(fr, (float, int)):
        v = v * fr
    else:
        v = fr[0](v)

    # Convert from the base unit
    if isinstance(to, (float, int)):
        v = v / to
    else:
        v = to[1](v)

    return v / tomul


def si_format_number(number: float, digits=2) -> str:
    """
    Format a number with SI prefixes. 1000 becomes 1k, etc
    """
    if number == 0:
        return "0"
    if number > 10**15:
        return str(iround(number / 1_000_000_000_000_000.0, digits)) + "P"
    if number > 10**12:
        return str(iround(number / 1_000_000_000_000.0, digits)) + "T"
    if number > 1_000_000_000:
        return str(iround(number / 1000_000_000.0, digits)) + "G"
    if number > 1_000_000:
        return str(iround(number / 1_000_000.0, digits)) + "M"
    if number > 1000:
        return str(iround(number / 1000.0, digits)) + "K"
    if number < 10**-12:
        return str(round(number * (10**-15), digits)) + "f"
    if number < 10**-9:
        return str(round(number * 1_000_000_000_000.0, digits)) + "p"
    if number < 10**-6:
        return str(iround(number * 1_000_000_000.0, digits)) + "n"
    if number < 0.001:
        return str(iround(number * 1_000_000.0, digits)) + "u"
    if number < 0.5:
        return str(iround(number * 1_000.0, digits)) + "m"
    return str(iround(number, digits))


time_as_seconds = {
    "year": 60 * 60 * 24 * 365,
    # A "month" as commonly used is a vauge unit. is it 28 days? 30? 31?
    # To solve that, I define it as 1/12th of a solar year.
    "month": 60 * 60 * 24 * 30.4368333333,
    "week": 60 * 60 * 24 * 7,
    "day": 60 * 60 * 24,
    "hour": 60 * 60,
    "minute": 60,
    "second": 1,
    "millisecond": 0.001,
    "microsecond": 0.000001,
    "nanosecond": 0.000000001,
    "picosecond": 0.000000000001,
    "femtosecond": 0.000000000000001,
}


time_as_seconds_abbr = {
    "yr": 60 * 60 * 24 * 365,
    # A "month" as commonly used is a vauge unit. is it 28 days? 30? 31?
    # To solve that, I define it as 1/12th of a solar year.
    "month": 60 * 60 * 24 * 30.4368333333,
    "w": 60 * 60 * 24 * 7,
    "d": 60 * 60 * 24,
    "h": 60 * 60,
    "m": 60,
    "s": 1,
    "ms": 0.001,
    "us": 0.000001,
    "ns": 0.000000001,
    "ps": 0.000000000001,
    "fs": 0.000000000000001,
}


def time_interval_from_string(s):
    """Take a string like '10 hours' or 'five minutes 32 milliseconds'
    or '1 year and 1 day' to a number of seconds"""
    regex = r"([0-9]*)\D*?(year|month|week|day|hour|minute|second|millisecond)s?"
    r = re.compile(regex)
    m = r.finditer(s)
    total = 0
    for i in m:
        multiplier = time_as_seconds[i.group(2).strip()]
        number = float(i.group(1))
        total += number * multiplier
    return total


def format_time_interval_long(t, max_units, clock=False):
    """Take a length of time t in seconds, and return a nicely formatted string
    like "2 hours, 4 minutes, 12 seconds".
    max_units is the maximum number of units to use in the string(7 will add a milliseconds field to times in years)

    """
    if clock:
        frac = t % 1
        t -= frac
        seconds = t % 60
        t -= seconds
        minutes = (t - (int(t / 3600) * 3600)) / 60
        t -= t % 3600
        hours = t / 3600

        s = "%02d:%02d" % (hours, minutes)
        if max_units > 2:
            s += ":%02d" % (seconds)
        if max_units > 3:
            # Adding 0.01 seems to help with some kind of obnoxious rounding bug thing. Prob a better way to do things.
            s += ":%03d" % (0.01 + frac * 1000)

        return s
    s = ""

    for i in sorted(time_as_seconds.items(), key=lambda x: x[1], reverse=True):
        if max_units == 0:
            return s[:-2]
        x = t % i[1]
        b = t - x
        y = (t - x) / i[1]

        t = t - b
        if y > 1:
            s += str(int(round(y))) + " " + i[0] + "s, "
            max_units -= 1
        elif y == 1:
            s += str(int(round(y))) + " " + i[0] + ", "
            max_units -= 1
    return s[:-2]


def format_time_interval_abbr(t, max_units, clock=False):
    """Take a length of time t in seconds, and return a nicely formatted string
    like "2 hours, 4 minutes, 12 seconds".
    max_units is the maximum number of units to use in the string(7 will add a milliseconds field to times in years)

    """
    if clock:
        frac = t % 1
        t -= frac
        seconds = t % 60
        t -= seconds
        minutes = (t - (int(t / 3600) * 3600)) / 60
        t -= t % 3600
        hours = t / 3600

        s = "%02d:%02d" % (hours, minutes)
        if max_units > 2:
            s += ":%02d" % (seconds)
        if max_units > 3:
            # Adding 0.01 seems to help with some kind of obnoxious rounding bug thing. Prob a better way to do things.
            s += ":%03d" % (0.01 + frac * 1000)

        return s
    s = ""
    for i in sorted(time_as_seconds_abbr.items(), key=lambda x: x[1], reverse=True):
        if max_units == 0:
            return s[:-1]
        x = t % i[1]
        b = t - x
        y = (t - x) / i[1]

        t = t - b
        if y > 1:
            if i[0] == "month":
                s += str(int(round(y))) + i[0] + "s "
            else:
                s += str(int(round(y))) + i[0] + " "
            max_units -= 1
        elif y:
            s += str(int(round(y))) + i[0] + " "
            max_units -= 1
    return s[:-1]


format_time_interval = format_time_interval_abbr


def str_to_int_si_multipliers(s):
    """Take a string of the form number[k|m|g] or just number and convert to an actual number
    '0'-> 0, '5k'->5000 etc. Does not do division!!!! m is mega not milli!!!"""
    s = s.lower()
    # This piece of code interprets a string like 89m or 50k as a number like 89 millon or 50,000
    if s.endswith("k"):
        return int(s[:-1]) * 1000
    elif s.endswith("m"):
        return int(s[:-1]) * 1000000
    elif s.endswith("g"):
        return int(s[:-1]) * 1000000000
    else:
        return int(s[:-1])


def iround(number, digits):
    if (number - int(number) == 0) or (digits == 0):
        return int(number)
    else:
        return round(number, digits)



# SPDX-FileCopyrightText: Copyright Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

"""This file deals with conversions between camelCase, snake_case, and kebab-case.
It can convert both strings and dict keys.
"""

# https://stackoverflow.com/a/44969381/2360612

from typing import Dict, Any
from functools import lru_cache


@lru_cache(maxsize=64)
def camel_to_kebab(s: str) -> str:
    return "".join(["_" + c.lower() if c.isupper() else c for c in s]).lstrip("_")


def kebab_to_snake(s: str):
    return s.replace("-", "_")


def snake_to_kebab(s: str):
    return s.replace("-", "_")


@lru_cache(maxsize=64)
def snake_to_camel(s: str):
    temp = s.split("_")
    return temp[0] + "".join(ele.title() for ele in temp[1:])


@lru_cache(maxsize=64)
def camel_to_snake(s: str):
    s2 = ""
    last = ""
    for i in s:
        if last.isalpha() and not last.isupper():
            if i.isupper():
                s2 += "_" + i.lower()
                continue
        last = i
        s2 += i.lower()
    return s2


@lru_cache(maxsize=64)
def any_to_snake(s: str):
    return camel_to_snake(kebab_to_snake(s))


@lru_cache(maxsize=64)
def any_to_camel(s: str):
    return snake_to_camel(kebab_to_snake(s))


@lru_cache(maxsize=64)
def any_to_kebab(s: str):
    return snake_to_kebab(camel_to_kebab(s))


def snakify_dict_keys(d: Dict[str, Any]) -> Dict[str, Any]:
    "Return a new dict with all keys converted to snake_case"
    return {camel_to_snake(kebab_to_snake(i)): d[i] for i in d}


def kebabify_dict_keys(d: Dict[str, Any]) -> Dict[str, Any]:
    "Return a new dict with all keys converted to kebab-case"
    return {snake_to_kebab(camel_to_kebab(i)): d[i] for i in d}


def camelify_dict_keys(d: Dict[str, Any]) -> Dict[str, Any]:
    "Return a new dict with all keys converted to camelCase"
    return {snake_to_camel(kebab_to_snake(i)): d[i] for i in d}



# SPDX-FileCopyrightText: Copyright 2015 Daniel Dunn
# SPDX-License-Identifier: LGPL-2.1-or-later

"""
Wraps the very simple scheduling module in a way that supports
repeating events, error reporting, and weakref-based cleanup.

Note that weakref cleanup is never a good idea to
rely on for correctness, it's just a tool to manage resources.

## Example
```python
# Set up logging, for demo purposes

import logging
import sys

logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

import time
import gc
import scullery.scheduling

# Direct decoration doesn't work on classes
@scullery.scheduling.every(1)
def f():
    print("test")

time.sleep(4)

# You could delete f and it would go away, but lets not because
# del f
f.unregister()


# Since you can't deco
class Foo():
    def __init__(self):
        # Must keep a reference to scheduled
        # Or it will stop
        self.scheduled = scullery.scheduling.every(self.bar,1)

    def bar(self):
        print("test 2")


f = Foo()
time.sleep(3)
# We don't call unregister, so we get a warning.
# f.scheduled.unregister()
del f
# Garbage collect so the deleter runs right away
gc.collect()
time.sleep(3)
```
"""

import threading
import sys
import time
import sched
import traceback
import logging
import types
from typing import Any, overload
from collections.abc import Callable

from scullery import workers, util


logger = logging.getLogger("system.scheduling")

localLogger = logging.getLogger("scheduling")


# These hooks are called on any exception in a function
# And passed the function
function_error_hooks: list[Callable[[Callable[..., Any]], Any]] = []


def handle_first_error(f: Callable):
    "Callback to deal with the first error from any given event"
    logger.exception(f"Error in {f}")


enumerate = enumerate


class BaseEvent:
    def __init__(self):
        self.exact = 0
        self.schedID = None


# Event API(not public):
# _schedule: if schedule is false calculate next runtime, insert self, set scheduled flag to true.
# only call under lock

# schedule: acquire lock, if so call _schedule

# _run: acquire lock(or do nothing if already running)
# do the actual action, and reschedule self if it's a repeating event

# run: use the worker pool to run an event, or run directly if fast enough(under 1ms)


class Event(BaseEvent):
    "Does function at time provided there is a strong referemce to f still by then"

    def __init__(self, function: Callable[[], Any], time: float):
        """_summary_

        Args:
            function (Callable[[],Any]): The function to call
            time (float): The time.time() at which to schedule.
        """
        BaseEvent.__init__(self)
        self.f = util.universal_weakref(function)
        self.fstr = str(function)
        self.time = time
        self.errored = False
        self.stopped = False

    def schedule(self):
        scheduler._insert(self)

    def run(self):
        workers.do(self._run)

    def _run(self):
        if self.stopped:
            return
        try:
            f = self.f()
            if not f:
                self.unregister()
                logger.warning(f"{self.fstr} was deleted before being called.")
            else:
                f()
        except Exception:
            if f:
                for i in function_error_hooks:
                    i(f)
        finally:
            del f

    def _unregister(self):
        scheduler.remove(self)

    # We want to use the worker pool to unregister so
    # that we know which thread the scheduler.unregister call is
    # going to be in to prevent deadlocks.
    # Also, we take a dummy var so we can use this as a weakref callback
    def unregister(self, dummy: Any = None):
        "Cancel running the event"
        self.stopped = True
        workers.do(self._unregister)


class BaseRepeatingEvent(BaseEvent):
    """Does function every interval seconds in real time,
    and stops if you don't keep a reference to function"""

    def __init__(
        self,
        function: Callable[[], Any],
        interval: float,
    ):
        """
        Args:
            function (Callable[[],Any]): Function to call
            interval (float): Interval
        """
        BaseEvent.__init__(self)
        self.f = util.universal_weakref(function)
        self.fstr = str(function)
        self.interval = float(interval)

        # True if the event is in the scheduler queue or the worker queue,
        # And should only be set under lock
        self.scheduled = False

        self.errored = False
        self.lock = threading.Lock()
        self.lastrun = 0.0
        # This flag is here to slove a really annoying problem.
        # If you unregister just before the reschedule function acquires the lock,
        # The object just gets rescheduled like nothing happened.
        # This stop function ensures they actually stop.
        self.stop = False
        del function

    def __repr__(self) -> str:
        try:
            f = self.f()
            if not f:
                f = f"{self.fstr}(dead)"
            f = str(f)
            return (
                "<BaseRepeatingEvent at "
                + str(id(self))
                + f
                + " every "
                + str(self.interval)
                + "s >"
            )
        except Exception:
            print(traceback.format_exc())
            return super().__repr__()

    def schedule(self):
        """Insert self into the scheduler. Not normally called
        by user code on repeating events, use register()

        Note for subclassing: The responsibility of this function
        to check if it is already scheduled, return if so,
        if not, if must reschedule itself by setting self.time
        to some time in the future.
        It must then call scheduler._insert(self)

        This must happen in a threadsafe and atomic way
        because the scheduler thread will call
        this every once in a while, just to be sure,
        in case an error occurred in the reschedule process

        We implement this at the moment by having a
        separate _schedule functinon for when we are already
        under self.lock
        """
        if self.lock.acquire(timeout=0.5):
            try:
                if self.scheduled:
                    return
                if not self.lastrun:
                    self.lastrun = time.time()
                self._schedule()
            finally:
                self.lock.release()
        else:
            logger.warning(
                f"Tried to schedule something that is still running: {str(self.f())}"
            )

    def _schedule(self):
        raise NotImplementedError()

    def register(self):
        """Register self in the list of
        repeating events to be automatically scheduled."""
        self.stop = False
        scheduler.register_repeating(self)
        self.schedule()

    def _unregister(self):
        scheduler.unregister(self)

    # We want to use the worker pool to unregister so that we know which thread the scheduler.unregister call is
    # going to be in to prevent deadlocks. Also, we take a dummy var so we can use this as a weakref callback
    def unregister(self, dummy=None):
        self.stop = True
        workers.do(self._unregister)

    def run(self):
        workers.do(self._run)

    def _run(self):
        # Safe to set outside lock I think. If there is
        # more than one scheduled copy there's a problem anyway.
        self.scheduled = False

        if self.stop:
            return

        # Don't run way too fast but do allow some tolerance in the speed.
        if time.time() - self.lastrun < (self.interval / 3):
            return

        self.lastrun = time.time()

        # We must have been pulled out of the event queue or we wouldn't be running.
        # If somehow there is another copy, exit and let recovery reschedule us later.

        if self.lock.acquire(timeout=1):
            try:
                f = self.f()
                if not f:
                    self.unregister()
                    logger.warning(f"{self.fstr} was automatically unregistered.")
                else:
                    f()
                # self._schedule()
            except Exception:
                if f:
                    for i in function_error_hooks:
                        i(f)

                if not self.errored:
                    try:
                        global handle_first_error
                        if f:
                            handle_first_error(f)
                    except Exception:
                        logging.exception(
                            "Error handling first error in repeating event"
                        )
                self.errored = True
            finally:
                # We have to reschedule no matter what.
                try:
                    self._schedule()
                except Exception:
                    # Don't even trust logging here. I'm being extremely paranoid about a deadlock.
                    try:
                        logging.exception("Error scheduling")
                    except Exception:
                        print(traceback.format_exc(6))
                self.lock.release()
                del f
                sys.last_traceback = None
        else:
            print(self.lock)


class RepeatingEvent(BaseRepeatingEvent):
    """Represents a repeating event that is not synced to
    multiples of wall clock time"""

    def __init__(self, *args, **kwargs):
        BaseRepeatingEvent.__init__(self, *args, **kwargs)

    def _schedule(self):
        """Calculate next runtime and put self into the queue.
        Should only ever be called under lock"""
        if self.scheduled:
            return

        # Don't alow unlimited amounts of winding up a big queue.
        t = max((self.lastrun + self.interval), ((time.time() + self.interval) - 5))
        self.time = t
        self.scheduled = True
        scheduler._insert(self)


class RepeatWhileEvent(RepeatingEvent):
    "Does function every interval seconds, and stops if you don't keep a reference to function"

    def __init__(self, function, interval):
        self.ended = False
        RepeatingEvent.__init__(self, function, interval)

    def _run(self):
        if self.ended:
            return
        if self.lock.acquire(False):
            try:
                f = self.f()
                if not f:
                    self.unregister()
                else:
                    r = f()
                    if not r:
                        self.unregister()
                        self.ended = True
            finally:
                self.lock.release()
                del f


class WrappedFunction:
    def __init__(self, f, e: RepeatingEvent):
        self.f = f
        self.e = e

    def __call__(self, *a, **k):
        return self.f(*a, **k)

    def unregister(self):
        self.e.unregister()


class NewScheduler:
    """
    represents a thread that constantly runs tasks which are objects having a time property that determins when
    their run method gets called. Inserted tasks use a lockless double buffered scheme.
    """

    def __init__(self):
        self._lock = threading.RLock()
        self._repeatingtasks = []
        self.daemon = True
        self.name = "SchedulerThread"
        self._lastrecheckedschedules = time.time()
        self._lf = time.time()
        self._running = False

        self._wakeUp = threading.Event()
        self._sched = sched.scheduler(timefunc=time.time, delayfunc=self.delay)

    def start(self):
        with self._lock:
            if self._running:
                return
            self._running = True
            self.thread = threading.Thread(
                daemon=self.daemon, target=self.run, name="schedulerthread"
            )
            self.thread2 = threading.Thread(
                daemon=self.daemon,
                target=self.manager,
                name="schedulerthread_errorrecovery",
            )

            self.thread.start()
            self.thread2.start()

    def every_second(self, f: Callable[[], Any]) -> WrappedFunction:
        """Run the callable f at the specified interval.
        Can be used as a decorator"""
        return self.every(f, 1)  # type: ignore

    def every_minute(self, f: Callable[[], Any]) -> WrappedFunction:
        return self.every(f, 60)  # type: ignore

    def every_hour(self, f: Callable[[], Any]) -> WrappedFunction:
        return self.every(f, 3600)  # type: ignore

    @overload
    def every(
        self, interval: float = 0
    ) -> Callable[[Callable[[], Any]], Callable[[], Any]]:
        pass

    @overload
    def every(self, f: Callable[[], Any], interval: float = 0) -> WrappedFunction:
        pass

    def every(self, f: Callable[[], Any] | float, interval: float = 0):  # type: ignore
        """Overloaded:
        every(interval) returns a decorator
        every(function, interval) calls the function

        The decorated or registered function will have an unregister() added,
        And it will also go away if you don't keep a reference to it.

        Call unregister() when you're done with it to avoid a logging warning
        when you delete it and it unregisters it for you.
        """

        if isinstance(f, (int, float)):
            if not interval == 0:
                raise ValueError(
                    "Must supply function and interval, or just interval for decorator."
                )

            def decorate(fn):
                return self.every(fn, f)

            return decorate

        if callable(f):
            if not interval:
                raise ValueError("Interval cannot be zero")

            interval = float(interval)
            e = RepeatingEvent(f, interval)
            e.register()
            if isinstance(f, types.MethodType):

                def f2():
                    f()

            else:
                f2 = f
            return WrappedFunction(f2, e)

    def schedule(self, f: Callable[[], Any], t: float, exact=False):
        t = float(t)
        e = Event(f, t)
        e.schedule()
        return e

    def schedule_repeating(self, f: Callable[..., Any], t: float, sync: bool = True):
        e = RepeatingEvent(f, float(t))
        e.register()
        return e

    def _insert(self, event, replaces=None):
        """Insert something that has a time  and a run
        property that wants its run called at time
        """

        if replaces:
            try:
                self._sched.cancel(replaces)
            except Exception:
                pass

        event.schedID = self._sched.enterabs(event.time, 1, event.run)
        # Only very fast events need to use this wake mechanism that burns CPU.
        # We have a min 0.15hz poll rate
        if event.time < (time.time() + 0.15):
            self._wakeUp.set()

    def remove(self, event):
        "Remove something that has a time and a run property that wants its run to be called at time"
        with self._lock:
            try:
                self._sched.cancel(event.schedID)
            except ValueError:
                pass
            except Exception:
                logger.exception("failed to remove event")

    def register_repeating(self, event):
        "Register a RepeatingEvent class"
        with self._lock:
            self._repeatingtasks.append(event)

    def unregister(self, event):
        "unregister a RepeatingEvent"
        with self._lock:
            try:
                try:
                    if event in self._repeatingtasks:
                        self._repeatingtasks.remove(event)
                except KeyError:
                    pass

                try:
                    self._sched.cancel(event.schedID)
                except ValueError:
                    pass
                except Exception:
                    logger.exception(
                        "failed to remove event, perhaps it was not actually scheduled"
                    )

            except Exception:
                logger.exception("failed to unregister event")

    def manager(self):
        while 1:
            time.sleep(30)
            with self._lock:
                self._do_error_recovery()

    # Custom delay func because we must be able to recieve new events while waiting
    def delay(self, t):
        self._wakeUp.clear()
        self._wakeUp.wait(min(t, 0.15))

    def run(self):
        while 1:
            try:
                self._sched.run()
            except Exception:
                logging.exception("Error in scheduler thread")

    def _do_error_recovery(self):
        for i in self._repeatingtasks:
            try:
                if not i.scheduled or time.time() - i.lastrun > (i.interval * 2):
                    # Give them 10 seconds to finish what they are doing and schedule themselves.
                    if i.lastrun < time.time() - 10:
                        # Let's maybe not block the entire scheduling thread
                        # If one event takes a long time to schedule or if it
                        # Is already running and can't schedule yet.

                        # On the off chance it actually IS scheduled, replace whatever was there last.
                        workers.do(i.schedule)
                        logger.debug(
                            "Rescheduled "
                            + str(i)
                            + "using error recovery, could indicate a bug somewhere, or just a long running event."
                        )
            except Exception:
                logger.exception("Exception while scheduling event")


scheduler = NewScheduler()
scheduler.start()

every = scheduler.every



# SPDX-FileCopyrightText: Copyright 2016 Daniel Dunn
# SPDX-License-Identifier: GPL-3.0-only
from __future__ import annotations

import time
import threading
import weakref

import beartype
from typing import Any
from collections.abc import Callable
from scullery import workers, util, scheduling

# import time
# import scullery.statemachines

# sm = scullery.statemachines.StateMachine("start")

# sm.add_state()

# sm.add_state('start', exit=lambda: print('Exiting start state'))
# sm.add_state('state2', enter=lambda: print('Entering state 2'))
# sm.add_state('state3', enter=lambda: print('Entering state 3'))

# sm.add_rule('start', 'my_event', 'state2')
# sm.set_timer('state2', 2, 'state3')

# # Event triggered, now we are in state2
# sm.event('my_event')

# time.sleep(3)

# # The timer went off, now we are in state 3
# # Stateage returns a tuple of state, age
# # Where age is how long we've been in that state.
# print(sm.stateage)


illegal_name_chars = "~!@#$%^&*()-=+`<>?,./;':\"\\[]{}\n\r\t "


def unboundProxy(self, f):
    def f2(*args, **kwargs):
        f(*args, **kwargs)

    return f2


def makechecker(ref: weakref.ref[StateMachine]):
    def timer_check():
        m = ref()
        if m:
            m._check_timer()

    return timer_check


def runSubscriber(f, state):
    f(state)


class UpdateControl:
    pass


class StateMachine:
    def __init__(self, start="start"):
        """
        Represents an State Machine or FSA
        Args:
            start (str, optional): _description_. The initial state. Defaults to "start".
        """

        self.states = {}
        self.state = start
        self.prev_state = None
        self.entered_state = time.time()
        # Used to ensure that if one leaves and reenters a state just as a timer is firing it does not trigger anything.
        self._transiton_count = 0
        self.lock = threading.RLock()

        # Subscribers, as lists of function weakrefs indexed by what state entrance they are subscribed to to
        self._subscribers: dict[str, list[weakref.ref[Callable]]] = {}

        # Used for skipping ahead to quickly test timers and things like that.
        self._time_offset = 0

        self.add_state("__closed__")
        # Placeholder state
        self.add_state(start)

    def __call__(self, event):
        "Trigger an event, return the current state"
        self.event(event)
        return self.state

    def __repr__(self):
        return "<State machine at %d in state %s, entered %d ago>" % (
            id(self),
            self.state,
            time.time() - self.entered_state,
        )

    def subscribe(self, f: Callable[[str], Any], state="__all__"):
        """Cause function f to be called when the machine enters the given state.
        If the state is __all__, causes
        f to be called whenever the state changes at all.
        Uses weak refs, so you must maintain a reference to f

        Args:
            f (_type_): The function
            state (str, optional): The specific state to subscribe to. Defaults to "__all__".
        """
        with self.lock:
            # First clean up old subscribers. This is slow to do thi every time, but it should be infrequent.
            for i in self._subscribers:
                self._subscribers[i] = [i for i in self._subscribers[state] if i()]
            self._subscribers = {
                i: self._subscribers[i]
                for i in self._subscribers
                if self._subscribers[i]
            }

            if state not in self._subscribers:
                self._subscribers[state] = []
            self._subscribers[state].append(util.universal_weakref(f))

    def unsubscribe(self, f: Callable[[str], Any], state="__all__"):
        """

        Args:
            f (_type_): The function
            state (str, optional): The specific state to unsub to. Defaults to "__all__".
        """
        with self.lock:
            for i in self._subscribers:
                self._subscribers[i] = [i for i in self._subscribers[state] if i()]
            self._subscribers = {
                i: self._subscribers[i]
                for i in self._subscribers
                if self._subscribers[i]
            }

            torm = None
            for i in self._subscribers[state]:
                if i() == f:
                    torm = i
            if torm:
                self._subscribers[state].remove(torm)

    @property
    def age(self):
        return time.time() - self.entered_state

    @property
    def stateage(self) -> tuple[str, float]:
        """_Get the state, and how long it's been in that state

        Returns:
            tuple[str, float]: The state and age of the state in seconds
        """
        with self.lock:
            return (self.state, time.time() - self.entered_state)

    def _check_timer(self):
        "Poll function for any timers on the state."
        with self.lock:
            if self.states[self.state].get("timer"):
                if (
                    (time.time() + self._time_offset) - self.entered_state
                ) > self.states[self.state]["timer"][0]:
                    # Get the destination
                    x = self.states[self.state]["timer"][1]

                    # If it's a function, call it to get the actual destination.
                    if isinstance(x, str):
                        self._goto(x)
                    else:
                        x = x(self)
                        if x:
                            self._goto(x)
                else:
                    self._configure_timer()

    def _configure_timer(self):
        "Sets up the timer. Needs to be called under lock"

        if hasattr(self, "schedulerobj"):
            self.schedulerobj.unregister()
            del self.schedulerobj

        # If for any reason we get here too early, let's just keep rescheduling
        if self.states[self.state].get("timer"):
            # If we haven't already passed the time of the timer
            if ((time.time() + self._time_offset) - self.entered_state) < self.states[
                self.state
            ]["timer"][0]:
                f = makechecker(util.universal_weakref(self))
                self.schedulerobj = scheduling.scheduler.schedule(f, time.time() + 0.08)

                # Keep a ref so it doesn't get GCed
                self.schedulerobj.func_ref = f  # type: ignore

            # If we have already passed that time, just do it now.
            # This is here for faster response when skipping ahead.
            else:
                workers.do(self._check_timer)

    def seek(self, t, condition=None):
        """
        Seek ahead to a given position in the curren state's timeline, but only if the"""
        with self.lock:
            if condition and (not condition == self.state):
                return
            pos = time.time() - self.entered_state
            self._time_offset = t - pos
            self._configure_timer()

    @beartype.beartype
    def add_state(
        self,
        name: str,
        rules: None | dict[str, str | Callable[[], str | None]] = None,
        enter: str | Callable | None = None,
        exit: str | Callable | None = None,
    ):
        """
        Create a new state.  Keys in rules must be event names,
        and values must be either state names or callables that return
        either a state name to go to, or None for no transition.

        Args:
            name (str): Name of the state. May not contain anything in illegal_name_chars
            rules (dict, optional): Dict of rules for the state
            enter (str | Callable | None, optional): Function to be called when entering
            exit (str | Callable | None, optional): Function to be called when exiting

        Raises:
            ValueError: _description_
        """
        for i in illegal_name_chars:
            if i in name:
                raise ValueError("Forbidden special character")
        with self.lock:
            self.states[name] = {
                "rules": rules or {},
                "enter": enter,
                "exit": exit,
                "conditions": [],
            }

    def set_timer(self, state: str, time: float | int, dest: str | Callable):
        """Add timer rule to a state.  When machine is in state continually
        for that many seconds, go to dest.

        Args:
            state (str): Starting state
            time (float | int): Timer duration
            dest (str | Callable): Dest state or callable returning name of dest
        """
        with self.lock:
            if dest:
                self.states[state]["timer"] = [time, dest]

    def remove_state(self, name):
        raise RuntimeError("Not supported now")

    @beartype.beartype
    def add_rule(self, start: str, event: str | Callable, to: str | Callable):
        """
        Add a rule to handle what should happen if event occurs while the machine is in state.
        The third parameter representing the destination state may also be a function.

        The function must take one parameter, the machine itself,
        and return either None for no Transiton or a string representing the new state.
        This lets you implement conditions and branchingself.
        """
        with self.lock:
            if isinstance(event, str):
                self.states[start]["rules"][event] = to
            elif callable(event):
                self.states[start]["conditions"].append((event, to))
                self._setupPolling()

    def del_rule(self, start, event):
        with self.lock:
            if event in self.states[start]["rules"]:
                del self.states[start]["rules"][event]
            elif event in self.states[start]["conditions"]:
                del self.states[start]["conditions"][event]
            else:
                raise KeyError("No such rule")

    def event(self, event):
        """Tell the machine that a specific event just occurred. If there is a matching Transiton rule for that event,
        then we do the current state's exit func, enter the new state, and do it's enter func"""
        with self.lock:
            if self.state not in self.states:
                return self.state

            s = self.states[self.state]

            if event in s["rules"]:
                x = s["rules"][event]
                # If the rule destination is a string, just use it. If it is not a string, then
                # It must be a function because those are the only two valid destination types.

                # If it's a function, call it to get the actual destination.
                if isinstance(x, str):
                    self._goto(x)
                else:
                    x = x(self)
                    if x:
                        self._goto(x)
            return self.event

    def check(self):
        "Check the function based condition rules"
        with self.lock:
            s = self.states.get(self.state, None)
            if s:
                # Check all the function rules, poll them all,
                # and should any happen to be true, we follow that rule
                # Like any other.
                for i in s["conditions"]:
                    if i[0]():
                        self._goto(i[1])

    def jump(self, state, condition=None):
        self.goto(state, condition)

    def goto(self, state, condition: str | None = None):
        """
        Jump to a specified state. If condition is not None,
        only jump if condition it matches the current state."
        """
        with self.lock:
            if condition and not self.state == condition:
                return
            self._goto(state)

    def close(self):
        """Stop any timers and go to the special state __closed__
        Do not use object after this.
        """
        self.goto("__closed__")
        if hasattr(self, "schedulerobj"):
            try:
                self.schedulerobj.unregister()
                del self.schedulerobj
            except Exception:
                pass

        if hasattr(self, "pollingscheduledfunction"):
            try:
                if self.pollingscheduledfunction:
                    self.pollingscheduledfunction.unregister()
                del self.pollingscheduledfunction
            except Exception:
                pass

    def __del__(self):
        self.close()

    def _setupPolling(self):
        "Start polling the function based events, if there are any to poll for"
        if hasattr(self, "pollingscheduledfunction"):
            if self.pollingscheduledfunction:
                self.pollingscheduledfunction.unregister()
            del self.pollingscheduledfunction

        if self.states[self.state].get("conditions"):
            self.pollingscheduledfunction = scheduling.scheduler.every(
                self.check, 1 / 24
            )

    def _goto(self, state):
        "Must be called under the lock"

        s = self.states[self.state]
        s2 = self.states[state]

        # Do the old state's exit function
        if s["exit"]:
            s["exit"]()

        self.prev_state = self.state
        self.state = state
        # Record the time that we entered the new state
        self.entered_state = time.time()
        self._configure_timer()

        self._time_offset = 0

        if hasattr(self, "schedulerobj"):
            self.schedulerobj.unregister()
            del self.schedulerobj

        if self.states[self.state].get("timer"):
            f = makechecker(util.universal_weakref(self))
            self.schedulerobj = scheduling.scheduler.schedule(
                f, time.time() + self.states[self.state].get("timer")[0]
            )

            # Keep a strong reference
            self.schedulerobj.func_ref = f  # type: ignore

        self._setupPolling()

        # Do the entrance function of the new state
        if s2["enter"]:
            s2["enter"]()

        # Handle the subscribers
        if state in self._subscribers:
            for i in self._subscribers[state]:
                x = i()
                if x:
                    runSubscriber(x, self.state)

        if "__all__" in self._subscribers:
            for i in self._subscribers["__all__"]:
                x = i()
                if x:
                    runSubscriber(x, self.state)

        # Increment the trans count. wrap at 2**64
        self._transiton_count = (self._transiton_count + 1) % 2**64
